classdef VariCamera < handle
    % VariCamera : a Pixel-wise varifocal camera model
    % coordinate ... C:C',  V:C, W:Cv or world co.   
    
	properties
        cam;
		% ROTATION MATRIX between WORLD & VCAMERA(C) COORDINATES
		W2V_R;
        V2W_R;
		% TRANSLATION
		W2V_t;
		V2W_t;
        % ROTATION MATRIX between VCAMERA(C) & CAMERA(C') COORDINATES
		V2C_R;
        C2V_R;
        % ROTATION MATRIX between WORLD & CAMERA(C') COORDINATES
		W2C_R;
        C2W_R;
		% TRANSLATION
		W2C_t;
		C2W_t;
        
		% INTRINSIC PARAMETERS
		f;          % focal length fc
        da;         % distance_air
        dg;         % glass(housing)
        n1=1.0;     % mu_a
        n2;         % mu_g
        n3=1.33;    % mu_w
        % FIELD OF VIEW (HALF)
        fovx;
        fovy;
        
        %Look up Table
        rg2IND; % r_g to the index of the look up table
        LookUP; % look up table

    end
    %METHODS
	methods
		function obj = VariCamera(dists, ns)
			if nargin ~= 2
				disp('[ERROR]CORRECT USAGE : VariCamera(dists, ns)');
            else
                if dists(1) obj.f  = dists(1); end
                if dists(2) obj.da = dists(2); end
                if dists(3) obj.dg = dists(3); end
                
                if ns(1) obj.n1 = ns(1); end
                if ns(2) obj.n2 = ns(2); end
                if ns(3) obj.n3 = ns(3); end
			end
        end
        %_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_
        % Backward projection pc -> pw 
        % Input : houing info & real scrn_pt & water z ...pc(r_pc, f), zqw 
        % Output: pt in water & virtual focal ...f_qw(), r_pc
        %/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
        function [vwv, fqw, pgv, qw] = backward(obj, pc, zqw)
            Rsize = size(pc(1,:));
            rpc = sqrt( pc(1,:).^2 + pc(2,:).^2);
            
            cosY = zeros(Rsize);
            sinY = zeros(Rsize);
            pcosY = pc(1,:)./rpc;
            psinY = pc(2,:)./rpc;
            cosY(rpc>0) = pcosY(rpc>0);
            sinY(rpc>0) = psinY(rpc>0);
            
            yaw = rpc/obj.f;
            rpa = obj.da *yaw;
            rva = (rpc./sqrt(rpc.*rpc + obj.f *obj.f *ones(Rsize)));
            rvg = (obj.n1/obj.n2)*rva;
            zvg = sqrt(1 - rvg.*rvg);
            rpg = rpa + (obj.dg *ones(Rsize)./zvg).*rvg;
            rvw = (obj.n1/obj.n3)*rva;
            zvw = sqrt(1 - rvw.*rvw);
            rpw = rpg + zqw.*rvw./zvw;
            fqw = zeros(Rsize);
            pfqw= (rpg./rvw).*zvw;
            fqw(rvw>0) = pfqw(rvw>0); 
            
            vwv = [rvw .*cosY; rvw .*sinY; zvw];
            pgv = [rpg .*cosY; rpg .*sinY; (obj.da +obj.dg)*ones(Rsize)];
            pwv = [rpw .*cosY; rpw .*sinY; (obj.da +obj.dg +zqw)*ones(Rsize)]; %
            
            qw = pwv;%obj.V2W_R * pwv + obj.V2W_t * ones(Rsize) ;
        end
        
         % function (for Newton method) 
         function F = Newton12(obj, y, rqw, zqw)
            F =  rqw -y* obj.da  -y* obj.dg./sqrt( (obj.n2^2-1)*y.^2 +obj.n2^2)...
                      -y.* zqw./sqrt( (obj.n3^2-1)*y.^2 +obj.n3^2);
         end
         
         % dF/dy 
         function F = diff_Newton12(obj, y, zqw)   
            F = - obj.da  -obj.dg./sqrt( (obj.n2^2-1)*y.^2 +obj.n2^2)...
                   - zqw./sqrt( (obj.n3^2-1)*y.^2 +obj.n3^2)...
                   + obj.dg*(obj.n2^2-1)*y.^2 ./sqrt( (obj.n2^2-1)*y.^2 +obj.n2^2).^3 ...
                   + zqw*(obj.n3^2-1).*y.^2 ./sqrt( (obj.n3^2-1)*y.^2 +obj.n3^2).^3;
         end
   
        %_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_
        % (1)Forward projection from water (All newton)
        % Input : houing info & world qw (x_qw, y_qw, z_qw) 
        % Output: real scrn_pt ...(up,vp), and virtual p_c
        %/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
        function [up,vp,p_c] = Water_forward(obj, q_w)
                setN = size(q_w(1,:));
                qw_v = obj.W2V_R * q_w + obj.W2V_t * ones(setN);
                rqw = sqrt(qw_v(1,:).^2 + qw_v(2,:).^2);
                zqw = qw_v(3,:) - (obj.da + obj.dg) * ones(setN);
                cosY = zeros(setN);
                sinY = zeros(setN);
                
                pcosY = qw_v(1,:)./rqw; % 
                psinY = qw_v(2,:)./rqw; % 
                
                cosY(rqw>0) = pcosY(rqw>0); % 
                sinY(rqw>0) = psinY(rqw>0); % 
  
                rqg0 = (obj.da +obj.dg) *rqw./((obj.da +obj.dg)*ones(setN) +zqw); 

                cnt=0;   
                yp = rqg0/(obj.da + obj.dg);
                while (sum(abs(obj.Newton12(yp,rqw,zqw)) < 1.0e-12) < setN(2) ) && (cnt < 30)
                    yp = yp - obj.Newton12(yp,rqw,zqw)./obj.diff_Newton12(yp,zqw);
                    cnt = cnt+1;    
                end
                
                    rpc = yp *obj.f;
                 %(rpc,) -> (xpc, ypc)
                    xpc = rpc.*cosY;
                    ypc = rpc.*sinY;
                    p_c = [ xpc; ypc; obj.f *ones(setN); ones(setN)];
                 %(xpc, ypc)v -> (u, v)
                    RTv2c = [obj.V2C_R, [0;0;0]];
                    UVs = obj.cam.InM * RTv2c * p_c;
                    up = UVs(1,:)./UVs(3,:);
                    vp = UVs(2,:)./UVs(3,:);
        end
        %_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_
        % (2)Forward projection from water k recursive + newton
        % Input : houing info & world qw (x_qw, y_qw, z_qw) 
        % Output: real scrn_pt ...(up,vp), and virtual p_c
        %/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
        function [up,vp,p_c] = Water_forward2(obj, q_w, k)
                setN = size(q_w(1,:));
                qw_v = obj.W2V_R * q_w + obj.W2V_t * ones(setN);
                rqw = sqrt(qw_v(1,:).^2 + qw_v(2,:).^2);
                zqw = qw_v(3,:) - (obj.da + obj.dg) * ones(setN);
                cosY = zeros(setN);
                sinY = zeros(setN);
                
                pcosY = qw_v(1,:)./rqw; % 
                psinY = qw_v(2,:)./rqw; % 
                
                cosY(rqw>0) = pcosY(rqw>0); % 
                sinY(rqw>0) = psinY(rqw>0); % 
              
                %perspect
                rqg0 = (obj.da +obj.dg) *rqw./((obj.da +obj.dg)*ones(setN) +zqw);  

                for rec = 1:k
                    index = rqg0 * obj.rg2IND;
                    fqw = obj.LookUP( floor(index)+1);
                    rqg0 = rqw.*fqw./(fqw+zqw);
                end
                
                tanvw = rqg0./fqw; 
                rvw = tanvw./sqrt(1+tanvw.^2);
                rva = obj.n3 * rvw;
                
                cnt=0;   
                yp = rva./sqrt(1-rva.^2);
                while (sum(abs(obj.Newton12(yp,rqw,zqw)) < 1.0e-12) < setN(2) ) && (cnt < 30)
                    yp = yp - obj.Newton12(yp,rqw,zqw)./obj.diff_Newton12(yp,zqw);
                    cnt = cnt+1;  
                end    
                    rpc = yp *obj.f;
                 %(rpc,) -> (xpc, ypc)
                    xpc = rpc.*cosY;
                    ypc = rpc.*sinY;
                    p_c = [ xpc; ypc; obj.f *ones(setN); ones(setN)];
                 %(xpc, ypc)v -> (u, v)
                    RTv2c = [obj.V2C_R, [0;0;0]];
                    UVs = obj.cam.InM * RTv2c * p_c;
                    up = UVs(1,:)./UVs(3,:);
                    vp = UVs(2,:)./UVs(3,:);
                 %F = [up;vp;p_c];
        end
         %_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_
        % (3)Forward projection from water (all recursive)
        % Input : houing info & world qw (x_qw, y_qw, z_qw) 
        % Output: real scrn_pt ...(up,vp), and virtual p_c
        %/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
        function [up,vp,p_c,v_w] = Water_forward3(obj, q_w)
                setN = size(q_w(1,:));
                qw_v = obj.W2V_R * q_w + obj.W2V_t * ones(setN);
                rqw = sqrt(qw_v(1,:).^2 + qw_v(2,:).^2);
                zqw = qw_v(3,:) - (obj.da + obj.dg) * ones(setN);
                cosY = zeros(setN);
                sinY = zeros(setN);
                
                pcosY = qw_v(1,:)./rqw; % 
                psinY = qw_v(2,:)./rqw; % 
                
                cosY(rqw>0) = pcosY(rqw>0); % 
                sinY(rqw>0) = psinY(rqw>0); % 

                %All recursive
                rqg0 = (obj.da +obj.dg) *rqw./((obj.da +obj.dg)*ones(setN) +zqw);   
                index = rqg0 * obj.rg2IND;
                fqw0 = obj.LookUP( floor(index)+1);
                rqg1 = rqg0;
                rqg0 = rqw.*fqw0./(fqw0+zqw);
                
                cnt=0;   
                 while (sum(abs(rqg0-rqg1) < 1.0e-12) < setN(2) ) && (cnt < 30)
                    rqg1 = rqg0;
                    index = rqg1 * obj.rg2IND;
                    fqw = obj.LookUP( floor(index)+1);
                    rqg0 = rqw.*fqw./(fqw+zqw);
                    cnt = cnt+1;
                 end
                tanvw = rqg0./fqw; 
                rvw = tanvw./sqrt(1+tanvw.^2);  
                rva = obj.n3 * rvw;
                v_w = [rvw.*cosY;rvw.*sinY;sqrt(1-rvw.^2)];
                yp = rva./sqrt(1-rva.^2);
                    rpc = yp *obj.f;
                 %(rpc,) -> (xpc, ypc)
                    xpc = rpc.*cosY;
                    ypc = rpc.*sinY;
                    p_c = [ xpc; ypc; obj.f *ones(setN); ones(setN)];
                 %(xpc, ypc)v -> (u, v)
                    RTv2c = [obj.V2C_R, [0;0;0]];
                    UVs = obj.cam.InM * RTv2c * p_c;
                    up = UVs(1,:)./UVs(3,:);
                    vp = UVs(2,:)./UVs(3,:);
        end
         
        function F = getC2V2W(obj)
              obj.C2V_R =  obj.V2C_R.';
              obj.V2W_R =  obj.W2V_R.';
              obj.V2W_t = -obj.V2W_R * obj.W2V_t;
              F = 1;
        end
        function F = getW2V2C(obj)
              obj.V2C_R =  obj.C2V_R.';
              obj.W2V_R =  obj.V2W_R.';
              obj.W2V_t = -obj.W2V_R * obj.V2W_t;
              F = 1;
        end
		 function F = updateBWC(obj)
              obj.W2C_R = obj.V2C_R * obj.W2V_R;
              obj.C2W_R = obj.V2W_R * obj.C2V_R;
              obj.W2C_t = obj.W2C_R * obj.W2V_t;
              obj.C2W_t = obj.V2W_t
              obj.fovx = atan(2.4/obj.f);
              obj.fovy = atan(1.8/obj.f);
              F = 1;
         end
         function F = clone(obj)
            F =  VariCamera([obj.f, obj.da, obj.dg], [1, obj.n2, obj.n3]);
            F.cam = obj.cam;
            F.W2V_R = obj.W2V_R;
            F.V2W_R = obj.V2W_R;
            F.W2V_t = obj.W2V_t;
            F.V2W_t = obj.V2W_t;
            F.V2C_R = obj.V2C_R;
            F.C2V_R = obj.C2V_R;
            F.updateBWC;
            F=1;
         end
         function F = getLookUp(obj,TABLE_NUM, CCD_SIZE)
            % getLookUp(TABLE_NUM, CCD_SIZE)
            %  table size *0.5 > sqrt(1280^2 + 960^2)
             flame_point_Oc = [CCD_SIZE*[1, 1, -1, -1; 1, -1, 1, -1]; obj.f, obj.f, obj.f, obj.f];
             flame_point = obj.C2V_R * flame_point_Oc;
             [vwv, fqw, pgv] = obj.backward( flame_point, 0);
             inf_rgv   = max( sqrt(pgv(1,:).^2 + pgv(2,:).^2) );
             obj.rg2IND = TABLE_NUM/inf_rgv;

             table_xgv = linspace(0, inf_rgv, TABLE_NUM);
             table_gv = [table_xgv; zeros(2,TABLE_NUM)];
             [u_p, v_p, p_c] = obj.Water_forward(table_gv);
             [vwv, obj.LookUP] = obj.backward( p_c, 0);
             F=1;
         end
    end
    
    
end

