clear all;
close all;

% enable profile viewer
profile on;


% load data %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
load('./data/sampleQw.mat');  % sample Qw (world coordinate), 6400 points
load('./data/vcamSim.mat');   % pixel-wise varifocal camera 
da = vcamSim.da;       % distance from camera center to housing
dg = vcamSim.dg;       % housing thickness
n2 = vcamSim.n2;       % refractive index of housing
n3 = vcamSim.n3;       % refractive index of water
f  = vcamSim.f;        % focal length
sampleN = size(Qw(1,:),2); % Num of samples (6400 from sampleQw.mat)

 
% Convert Qw for Agrawal2012 (SolveForwardProjectionCase3.m require points
% in the original camera coordinate system)

Qw_Oc = vcamSim.W2C_R * Qw + vcamSim.W2C_t*ones(size(Qw(1,:)));
og_Oc = vcamSim.V2C_R * [0;0;da+dg];
n_vct = -og_Oc / norm(og_Oc);


% Test %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
% Forward projection performance evaluation
% (0) Agrawal2012 solving 12-th polynomial (SolveForwardProjectionCase3.m)
% (1) ALL NEWTON              (pixel-wise varifocal camera) 
% (2)  p  RECURSIVE+NEWTON    (pixel-wise varifocal camera) 
% (3) ALL RECURSIVE           (pixel-wise varifocal camera) 
%_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/


% preallocate arrays to store the results
p_a_Oc = zeros(3, sampleN);
p_c = zeros(4, sampleN);
Img0 = zeros(3, sampleN);
Img1 = zeros(3, sampleN);
Img2 = zeros(3, sampleN);
Img3 = zeros(3, sampleN);


%%%% Case 0 %%%%
tic;
for k = 1 : sampleN
	p_a_Oc(:,k) = SolveForwardProjectionCase3(da, da+dg, n_vct, [n2; n3], Qw_Oc(:,k));
end
cost0 = toc;

% Convert to pixels (SolveForwardProjectionCase3 returns the position on the inner housing surface)
% This is a simple perspective projection.
% We use this as the ground-truth to evaluate the accuracy of our projection
Img0 = vcamSim.cam.InM * p_a_Oc;
Img0(1,:) = Img0(1,:)./Img0(3,:);
Img0(2,:) = Img0(2,:)./Img0(3,:);
Img0(3,:) = Img0(3,:)./Img0(3,:);


%%%% Case 1 %%%%
tic;
for k = 1 : sampleN
	[Img1(1,k),Img1(2,k),p_c(:,k)] = vcamSim.Water_forward(Qw(:,k));
end
cost1 = toc;


%%%% Case 2 %%%%
p_c = zeros(4, sampleN);
num_of_recurr = 2;
tic;
for k = 1 : sampleN
	[Img2(1,k),Img2(2,k),p_c(:,k)] = vcamSim.Water_forward2(Qw(:,k), num_of_recurr);
end
cost2 = toc;


%%%% Case 3 %%%%
p_c = zeros(4, sampleN);
tic;
for k = 1 : sampleN
	[Img3(1,k),Img3(2,k),p_c(:,k)] = vcamSim.Water_forward3(Qw(:,k));
end
cost3 = toc;


%_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
% without using for loop
%_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/


%%%% Case 1 without loop %%%%
tic;
p_c = zeros(4, sampleN);
	[Img3(1,:),Img3(2,:),p_c] = vcamSim.Water_forward(Qw);
cost1_ = toc;


%%%% Case 2 without loop %%%%
tic;
p_c2 = zeros(4, sampleN);
	[Img3(1,:),Img3(2,:),p_c2] = vcamSim.Water_forward2(Qw, 4); % Please Change last input arg
cost2_ = toc;


%%%% Case 3 without loop %%%%
tic;
p_c3 = zeros(4, sampleN);
	[Img3(1,:),Img3(2,:),p_c3] = vcamSim.Water_forward3(Qw);
cost3_ = toc;


% Results %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Running time
fprintf('Average costs of %d forward projections:\n\n', sampleN);
fprintf('  by Agrawal 2012 (with for loop)           = %f sec\n', cost0/sampleN);
fprintf('  by PVCM-Newton (with for loop)            = %f sec\n', cost1/sampleN);
fprintf('  by PVCM-Recurrrent-Newton (with for loop) = %f sec\n', cost2/sampleN);
fprintf('  by PVCM-Recurrrent (with for loop)        = %f sec\n', cost3/sampleN);
fprintf('  by PVCM-Newton (w/o for loop)             = %f sec\n', cost1_/sampleN);
fprintf('  by PVCM-Recurrrent-Newton (w/o for loop)  = %f sec\n', cost2_/sampleN);
fprintf('  by PVCM-Recurrrent (w/o for loop)         = %f sec\n', cost3_/sampleN);
fprintf('\n');

% Errors
err1s = abs(Img0(1:2,:) - Img1(1:2,:));
err1  = max( sqrt(err1s(1,:).^2 + err1s(2,:).^2) );
err2s = abs(Img0(1:2,:) - Img2(1:2,:));
err2  = max( sqrt(err2s(1,:).^2 + err2s(2,:).^2) );
err3s = abs(Img0(1:2,:) - Img3(1:2,:));
err3  = max( sqrt(err3s(1,:).^2 + err3s(2,:).^2) );

fprintf('Maximum projection errors in comparison with Agrawal 2012\n\n');
fprintf('  by PVCM-Newton                            = %f px\n', err1);
fprintf('  by PVCM-Recurrrent-Newton                 = %f px\n', err2);
fprintf('  by PVCM-Recurrrent                        = %f px\n', err3);
fprintf('\n');


% Visualization %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

RTc2w = [vcamSim.C2W_R, vcamSim.C2W_t];
RTv2w = [vcamSim.V2W_R, vcamSim.V2W_t];
RTw2w = [1,0,0,0;0,1,0,0;0,0,1,0];
fovx = atan(2.4/f);
fovy = atan(1.8/f);
Flms = [ [-tan(fovx); -tan(fovy); 1], ...
         [ tan(fovx); -tan(fovy); 1], ...
         [ tan(fovx); +tan(fovy); 1], ...
         [-tan(fovx); +tan(fovy); 1] ];
% camera 
Co_o = RTc2w * [0;0;0;1];
Flm = RTc2w * [5*f*Flms; [1,1,1,1]];
Clines = [Co_o,Flm(:,1),Flm(:,2),Co_o,Flm(:,3),Flm(:,4),Co_o];
line(Clines(1,:), Clines(2,:), Clines(3,:));  
Clines = [Flm(:,2),Flm(:,3),Flm(:,4),Flm(:,1) ];
line(Clines(1,:), Clines(2,:), Clines(3,:));

Flm = [da*Flms; [1,1,1,1]];
Flm(3,:) = Flm(3,:) -da;
Clines = [Flm(:,1),Flm(:,2),Flm(:,3),Flm(:,4),Flm(:,1)];
line(Clines(1,:), Clines(2,:), Clines(3,:));

Flm = [(da+dg)*Flms; [1,1,1,1]];
Flm(3,:) = Flm(3,:) -da;
Clines = [Flm(:,1),Flm(:,2),Flm(:,3),Flm(:,4),Flm(:,1)];
line(Clines(1,:), Clines(2,:), Clines(3,:));

hold all
scatter3(Qw(1,:),Qw(2,:),Qw(3,:));
UVs = RTc2w * [5 * p_c(1:3,:); p_c(4,:)];
scatter3(UVs(1,:),UVs(2,:),UVs(3,:));
grid on;

